/*
 * ipcrypt
 *
 * This source is licensed under the GPL Version 2. A few routines are
 * shamelessly ripped from the Linux kernel as stated below. All other
 * code is (C) 2009 Andreas Steinmetz (ast@domdv.de).
 */

/* system headers */

#include <linux/netfilter.h>
#include <libnetfilter_queue/libnetfilter_queue.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <pwd.h>
#include <grp.h>
#include <sched.h>
#include <stdlib.h>
#include <signal.h>
#include <time.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#define __USE_GNU
#include <unistd.h>
#undef _POSIX_SOURCE
#include <sys/syscall.h>
#include <sys/prctl.h>
#include <sys/capability.h>

/* crypto headers */

#include "common.h"
#include "md5.h"
#include "blowfish.h"

/* default values */

#define DFTKEY "/etc/ipcrypt.key"	/* default key file pathname	 */
#define DFTSCR "/etc/ipcrypt.sh"	/* default script file pathname	 */
#define DFTPID "/var/run/ipcrypt.pid"	/* default pid file pathname	 */
#define DFTTX  "1"			/* default tx nfqueue number	 */
#define DFTRX  "2"			/* default rx nfqueue number	 */

/* key file */

#define RANDOMDEV "/dev/urandom"	/* random device for key file	 */
#define KEYLENGTH 256			/* key file length		 */

/* global variables */

static char *pidfile=DFTPID;		/* pid file pathname		 */
static char *keyfile=DFTKEY;		/* key file pathname		 */
static char *script=DFTSCR;		/* up/down script pathname	 */
static char *txqnam=DFTTX;		/* tx nfqueue number		 */
static char *rxqnam=DFTRX;		/* rx nfqueue number		 */
static struct nfq_handle *h=NULL;	/* netfilter socket handle	 */
static struct nfq_q_handle *qh1=NULL;	/* netfilter tx queue handle	 */
static struct nfq_q_handle *qh2=NULL;	/* netfilter rx queue handle	 */
static int ipt=0;			/* iptables setup done flag	 */
static unsigned int seq=0;		/* ip id sequence counter	 */
static int usetime=0;			/* replay prevention mode enable */
static int hmacmode=0;			/* authentication mode enable	 */
static int txtcp=1;			/* tx tcp checksum create flag	 */
static int rxtcp=1;			/* rx tcp checksum create flag	 */
static int txudp=1;			/* tx udp checksum create flag	 */
static int rxudp=1;			/* rx udp checksum create flag	 */
static MD5HMDATA hmac;			/* hmac key used for IV gen.	 */
static MD5HMDATA hmac2;			/* hmac key used for auth	 */
static BLOWFISH_SCHEDULE(sched);	/* blowfish cipher key schedule	 */
static WORD08 white[256];		/* whitening buffer		 */
static char bfr[69632];			/* netfilter message buffer (68k)*/

/* inline assembler for x86 and x86_64 */

#if defined(__x86_64__) || defined(__i386__)

/*
 * ipcsum
 *
 * input:  p   - the pointer to the IP header
 *	   len - the size of the IP header
 *
 * return: the checksum in network byte order
 *
 * description: This routine creates the IP header checksum. It was
 *		shamelessly ripped from the Linux kernel.
 */

static inline unsigned short ip_csum(unsigned char *p,int len)
{
	unsigned int sum;

	asm("  movl (%1), %0\n"
	    "  subl $16, %2\n"
	    "  jbe 2f\n"
	    "  addl 4(%1), %0\n"
	    "  adcl 8(%1), %0\n"
	    "  adcl 12(%1), %0\n"
	    "1: adcl 16(%1), %0\n"
	    "  lea 4(%1), %1\n"
	    "  subl $4, %2\n"
	    "  jne      1b\n"
	    "  adcl $0, %0\n"
	    "  movl %0, %2\n"
	    "  shrl $16, %0\n"
	    "  addw %w2, %w0\n"
	    "  adcl $0, %0\n"
	    "  notl %0\n"
	    "2:"
	    : "=r" (sum), "=r" (p), "=r" (len)
	    : "1" (p), "2" (len)
	    : "memory");
	return (unsigned short)sum;
}

/*
 * csum_fold
 *
 * input:  sum - the intermediate checksum
 *
 * return: the checksum in network byte order
 *
 * description:	This routine folds an intermediate checksum into a
 *		16 bit checksum in network byte order. It was
 *		shamelessly ripped from the Linux kernel.
 */

static inline unsigned short csum_fold(unsigned int sum)
{
        asm("  addl %1,%0\n"
            "  adcl $0xffff,%0"
            : "=r" (sum)
            : "r" ((unsigned int)sum << 16),
              "0" ((unsigned int)sum & 0xffff0000));
        return (unsigned short)(~(unsigned int)sum >> 16);
}

/*
 * csum_nofold
 *
 * input:  saddr - the source address in network byte order
 *	   daddr - the destination addres in network byte order
 *	   len   - the tcp or udp datagram length
 *	   proto - the ip protocol
 *	   sum   - the intermediate checksum of the udp or tcp datagram
 *
 * return: an intermediate checksum
 *
 * description:	This routine adds the required header fields to
 *		a tcp or udp data checksum and returns an
 *		intermediate checksum. It was shamelessly ripped
 *		from the Linux kernel.
 */

static inline unsigned int csum_nofold(unsigned int saddr,
	unsigned int daddr,unsigned short len,unsigned short proto,
	unsigned int sum)
{
        asm("  addl %1, %0\n"
            "  adcl %2, %0\n"
            "  adcl %3, %0\n"
            "  adcl $0, %0\n"
            : "=r" (sum)
            : "g" (daddr), "g" (saddr),
              "g" ((len+proto)<<8), "0" (sum));
        return sum;
}

/*
 * memory_csum
 *
 * input:  d   - the pointer to the memory area
 *	   len - the length of the memory area
 *
 * return: an intermediate checksum
 *
 * description:	This routine creates an intermediate checksum of
 *		the speciified memory area.
 */

static inline unsigned int memory_csum(unsigned char *d,int len)
{
	unsigned int sum=0;

	asm("  movl %2,%%esi\n"
	    "  shrl $4,%2\n"
	    "  jz chksum2l%=\n"
	    "  clc\n"
	    "chksum1l%=:\n"
	    "  adcl (%1),%0\n"
	    "  adcl 4(%1),%0\n"
	    "  adcl 8(%1),%0\n"
	    "  adcl 12(%1),%0\n"
	    "  decl %2\n"
	    "  lea 16(%1),%1\n"
	    "  jnz chksum1l%=\n"
	    "  adcl $0,%0\n"
	    "chksum2l%=:\n"
	    "  andl $15,%%esi\n"
	    "  jz chksum6l%=\n"
	    "  movl %%esi,%2\n"
	    "  shrl $2,%2\n"
	    "  jz chksum4l%=\n\n"
	    "  clc\n"
	    "chksum3l%=:\n"
	    "  adcl (%1),%0\n"
	    "  decl %2\n"
	    "  lea 4(%1),%1\n"
	    "  jnz chksum3l%=\n"
	    "  adcl $0,%0\n"
	    "chksum4l%=:\n"
	    "  andl $3,%%esi\n"
	    "  jz chksum6l%=\n"
	    "  bt $1,%%esi\n"
	    "  jnc chksum5l%=\n"
	    "  movzwl (%1),%2\n"
	    "  lea 2(%1),%1\n"
	    "  addl %2,%0\n"
	    "  adcl $0,%0\n"
	    "chksum5l%=:\n"
	    "  bt $0,%%esi\n"
	    "  jnc chksum6l%=\n"
	    "  movzbl (%1),%2\n"
	    "  addl %2,%0\n"
	    "  adcl $0,%0\n"
	    "chksum6l%=:\n"
	    : "+r" (sum), "+r" (d), "+r" (len)
	    :
	    : "%esi"
	);
	return sum;
}

/*
 * csum
 *
 * input:  i   - a pointer to the ip header
 *	   d   - a pointer to a tcp or udp datagram
 *	   len - the datagram length
 *
 * return: a checksum in network byte order
 *
 * description:	This routine creates a checksum for udp or tcp datagrams.
 */

static inline unsigned short csum(struct iphdr *i,unsigned char *d,int len)
{
	return csum_fold(csum_nofold(i->saddr,i->daddr,len,i->protocol,
		memory_csum(d,len)));
}

/* slow generic code for all other platforms */

#else

/*
 * ipcsum
 *
 * input:  p   - the pointer to the IP header
 *	   len - the size of the IP header
 *
 * return: the checksum in network byte order
 *
 * description: This routine creates the IP header checksum.
 */

static inline unsigned short ip_csum(unsigned char *p,int len)
{
	unsigned long sum;	/* checksum		*/
	int j;			/* counter and index	*/


	/* for the whole IP header */

	for(sum=0,j=0;j<len;)
	{
		/* add current 16 bit to checksum */

		sum+=p[j++];
		sum+=((unsigned long)(p[j++]))<<8;
	}

	/* fold checksum */

	while(sum&0xffff0000)sum=(sum>>16)+(sum&0xffff);

	/* return checksum */

	return (unsigned short)(~sum);
}

/*
 * csum
 *
 * input:  i   - a pointer to the ip header
 *	   d   - a pointer to a tcp or udp datagram
 *	   len - the datagram length
 *
 * return: a checksum in network byte order
 *
 * description:	This routine creates a checksum for udp or tcp datagrams.
 */

static inline unsigned short csum(struct iphdr *i,unsigned char *d,int len)
{
	unsigned long sum;	/* checksum				*/
	int j;			/* counter and index			*/
	unsigned short clen;	/* datagram length in network byte order*/


	/* for the whole datagram */

	for(clen=htons(len),sum=0,j=0;j<len;)
	{
		/* add current 16(8) bits to checksum */

		sum+=d[j++];
		if(j==len)break;
		sum+=((unsigned long)(d[j++]))<<8;
	}

	/* add ip source address to checksum */

	d=(unsigned char *)(&i->saddr);
	sum+=d[0];
	sum+=((unsigned long)(d[1]))<<8;
	sum+=d[2];
	sum+=((unsigned long)(d[3]))<<8;

	/* add ip destination address to checksum */

	d=(unsigned char *)(&i->daddr);
	sum+=d[0];
	sum+=((unsigned long)(d[1]))<<8;
	sum+=d[2];
	sum+=((unsigned long)(d[3]))<<8;

	/* add datagram length to checksum */

	d=(unsigned char *)(&clen);
	sum+=d[0];
	sum+=((unsigned long)(d[1]))<<8;

	/* add protocol to checksum */

	sum+=((unsigned long)(i->protocol))<<8;

	/* fold checksum */

	while(sum&0xffff0000)sum=(sum>>16)+(sum&0xffff);

	/* return checksum */

	return (unsigned short)(~sum);
}

#endif

/*
 * encrypt
 *
 * input: hmdata - pointer tot the data for IV generation
 *	  hlen   - length of the data for IV generation
 *	  total  - length of the data to be encrypted
 *
 * inout: the unencrypted data on input, the encrypted data on output
 *
 * description:	This routine encrypts the given data using the specified
 *		data for IV generation.
 */

static inline void encrypt(void *hmdata,int hlen,unsigned char *data,int total)
{
	register WORD32 i;	/* general purpose usage		     */
	register WORD08 *ptr;	/* buffer fast access			     */
	register WORD08 w;	/* whitening access index		     */
	MD5(bfr);		/* hmac output buffer for IV/whitening setup */


	/* done if nothing to encrypt */

	if(!total)return;

	/* create IV and whitening base index from IV input data */

	md5hmac((unsigned char *)hmdata,hlen,bfr,&hmac);
	w=bfr[BLOWFISH_BLOCKSIZE];

	/* for all large blocks use CFB64 */

	while(total>=BLOWFISH_BLOCKSIZE)
	{
		/* encrypt buffer */

		blowfish_encrypt(bfr,sched);

		/* XOR data bytes with buffer, memorize new iv
		   and do whitening */

		for(i=0,ptr=bfr;i<BLOWFISH_BLOCKSIZE;i++)
		{
			*data^=*ptr;
			*ptr++=*data;
			*data++^=white[w++];
		}

		/* adjust data count */

		total-=BLOWFISH_BLOCKSIZE;
	}

	/* for all remaining data bytes use CFB8 */

	while(total--)
	{
		/* encrypt buffer */

		blowfish_encrypt(bfr,sched);

		/* XOR data byte with buffer */

		*data^=*bfr;

		/* shift iv left and append cipher byte and do whitening */

		for(i=0,ptr=bfr+1;i<BLOWFISH_BLOCKSIZE-1;i++)bfr[i]=*ptr++;
		bfr[i]=*data;
		*data++^=white[w++];
	}
}

/*
 * decrypt
 *
 * input: hmdata - pointer tot the data for IV generation
 *	  hlen   - length of the data for IV generation
 *	  total  - length of the data to be decrypted
 *
 * inout: the encrypted data on input, the unencrypted data on output
 *
 * description:	This routine decrypts the given data using the specified
 *		data for IV generation.
 */

static inline void decrypt(void *hmdata,int hlen,unsigned char *data,int total)
{
	register WORD32 i;	/* general purpose usage		     */
	register WORD08 *ptr;	/* buffer fast access			     */
	register WORD08 mem;	/* general purpose memory		     */
	register WORD08 w;	/* whitening access index		     */
	MD5(bfr);		/* hmac output buffer for IV/whitening setup */


	/* done if nothing to encrypt */

	if(!total)return;

	/* create IV and whitening base index from IV input data */

	md5hmac((unsigned char *)hmdata,hlen,bfr,&hmac);
	w=bfr[BLOWFISH_BLOCKSIZE];

	/* for all large blocks use CFB64 */

	while(total>=BLOWFISH_BLOCKSIZE)
	{
		/* encrypt buffer */

		blowfish_encrypt(bfr,sched);

		/* unwhite, memorize new iv and then XOR data bytes
		   with buffer */

		for(i=0,ptr=bfr;i<BLOWFISH_BLOCKSIZE;i++)
		{
			*data^=white[w++];
			mem=*data;
			*data++^=*ptr;
			*ptr++=mem;
		}

		/* adjust data count */

		total-=BLOWFISH_BLOCKSIZE;
	}

	/* for all remaining data bytes use CFB8 */

	while(total--)
	{
		/* encrypt buffer */

		blowfish_encrypt(bfr,sched);

		/* unwhite, memorize and then XOR data byte with buffer */

		*data^=white[w++];
		mem=*data;
		*data++^=*bfr;

		/* shift iv left and append cipher MSB */

		for(i=0,ptr=bfr+1;i<BLOWFISH_BLOCKSIZE-1;i++)bfr[i]=*ptr++;
		bfr[i]=mem;
	}
}

/*
 * tx
 *
 * input:  see netfilter documentation, if there is any
 *
 * return: see netfilter documentation, if there is any
 *
 * description:	This routine is the netfilter nfqueue callback used for
 *		encryption.
 */

static int tx(struct nfq_q_handle *qh,struct nfgenmsg *nfmsg,
	struct nfq_data *nfa,void *data)
{
	int id;				/* netfilter packet id		*/
	int len;			/* tptal datagram length	*/
	int ilen;			/* ip header length		*/
	int dlen;			/* tcp/udp total length		*/
	int doff;			/* tcp/udp payload offset	*/
	struct nfqnl_msg_packet_hdr *ph;/* netfilter packet header	*/
	union
	{
		struct iphdr *iph;	/* ip header			*/
		char *pkt;		/* ip header as char		*/
		unsigned char *bfr;	/* ip header as unsigned char	*/
	}u;
	union
	{
		unsigned char *bfr;	/* tcp/udp header as uns. char	*/
		struct udphdr *udp;	/* udp header			*/
		struct tcphdr *tcp;	/* tcp header			*/
	}v;
	struct
	{
		unsigned int timestamp;	/* replay protect timestamp or 0*/
		unsigned short id;	/* IP ID of datagram		*/
		unsigned short len;	/* datagram total length	*/
	}w;
	union
	{
		MD5(auth);		/* hmac output as bytes		*/
		unsigned short data[8];	/* hmac output as 16 bit values	*/
	}x;
	MD5DATA ctx;			/* hmac working context		*/


	/* try to get the netfilter packet id or use 0 if not possible */

	if((ph=nfq_get_msg_packet_hdr(nfa)))id=ntohl(ph->packet_id);
	else id=0;

	/* get packet, ignore packet if it is too short */

	if((len=nfq_get_payload(nfa,&u.pkt))<sizeof(struct iphdr)||
		len<(ilen=(u.iph->ihl<<2)))goto fail;

	/* ignore the packet if specified datagram length is larger than
	   the received packet length */

	if(len<ntohs(u.iph->tot_len))goto fail;

	/* ignore all fragmented packets */

	if(ntohs(u.iph->frag_off)&(IP_MF|IP_OFFMASK))goto fail;

	/* get pointer to protocol specific datagram part, get length
	   of protocol specific part */


	v.bfr=u.bfr+ilen;
	dlen=len-ilen;

	/* act according to protocol */

	switch(u.iph->protocol)
	{
	/* tcp */

	case IPPROTO_TCP:

		/* skip packet if protocol header is too short,
		   calculate payload offset from tcp header start */

		if(dlen<sizeof(struct tcphdr))goto fail;
		doff=v.tcp->doff<<2;
		if(dlen<doff)goto fail;
		break;

	/* udp */

	case IPPROTO_UDP:

		/* skip packet if protocol header is too short */

		if(dlen<sizeof(struct udphdr)||dlen<ntohs(v.udp->len))goto fail;

		/* set up udp payload offset from udp header start */

		doff=sizeof(struct udphdr);
		break;

	/* skip all other protocols */

	default:goto fail;
	}

	/* set up timestamp according to replay protection mode */

	if(!usetime)w.timestamp=0;
	else w.timestamp=((unsigned int)time(NULL))>>8;

	/* in case of a set DF bit */

	if(ntohs(u.iph->frag_off)&IP_DF)
	{
		/* use our own sequence counter for the IP ID */

		u.iph->id=htons(seq++);

		/* continue according to the replay protection mode */

		if(usetime)goto timecsum;
		goto ipcsum;
	}

	/* if replay protection is enabled */

	else if(usetime)
	{
		/* add the 2 timestamp LSB as top bits to IP TOS */

timecsum:	u.iph->tos&=0x3f;
		u.iph->tos|=(unsigned char)(w.timestamp<<6);

		/* generate new ip checksum */

ipcsum:		u.iph->check=0;
		u.iph->check=ip_csum(u.bfr,ilen);
	}

	/* make timestamp endian independent */

	w.timestamp=htonl(w.timestamp);

	/* copy IP ID and IP datagram length */

	w.id=u.iph->id;
	w.len=u.iph->tot_len;

	/* encrypt payload */

	encrypt(&w,sizeof(w),v.bfr+doff,dlen-doff);

	/* act according to protocol */

	switch(u.iph->protocol)
	{
	/* tcp */

	case IPPROTO_TCP:

		/* if authentication is enabled */

		if(hmacmode)
		{
			/* reset tcp checksum */

			v.tcp->check=0;

			/* create hmac of header and encrypted payload,
			   as well as the IV source data */

			md5hminit(&ctx,&hmac2);
			md5hmnext(v.bfr,dlen,&ctx);
			md5hmnext((unsigned char *)&w,sizeof(w),&ctx);
			md5hmend(x.auth,&ctx,&hmac2);

			/* copy first two bytes of hmac to checksum field */

			v.tcp->check=x.data[0];

			/* done */

			break;
		}

		/* done if transmit tcp checksum generation is disabled */

		if(!txtcp)break;

		/* update tcp checksum */

		v.tcp->check=0;
		v.tcp->check=csum(u.iph,v.bfr,dlen);
		break;

	/* udp */

	case IPPROTO_UDP:

		/* if authentication is enabled */

		if(hmacmode)
		{
			/* reset udp checksum */

			v.udp->check=0;

			/* create hmac of header and encrypted payload,
			   as well as the IV source data */

			md5hminit(&ctx,&hmac2);
			md5hmnext(v.bfr,dlen,&ctx);
			md5hmnext((unsigned char *)&w,sizeof(w),&ctx);
			md5hmend(x.auth,&ctx,&hmac2);

			/* copy first two bytes of hmac to checksum field */

			v.udp->check=x.data[0];

			/* done */

			break;
		}

		/* done if there is no udp sum or udp summing is disabled */

		if(!v.udp->check||!txudp)break;

		/* update udp checksum */

		v.udp->check=0;
		v.udp->check=csum(u.iph,v.bfr,dlen);
		if(!v.udp->check)v.udp->check=0xffff;
		break;
	}

	/* return modified packet to kernel */

	return nfq_set_verdict(qh,id,NF_ACCEPT,len,u.bfr);

	/* tell kernel to pass the unaltered packet */

fail:	return nfq_set_verdict(qh,id,NF_ACCEPT,0,NULL);
}

/*
 * rx
 *
 * input:  see netfilter documentation, if there is any
 *
 * return: see netfilter documentation, if there is any
 *
 * description:	This routine is the netfilter nfqueue callback used for
 *		decryption.
 */

static int rx(struct nfq_q_handle *qh,struct nfgenmsg *nfmsg,
	struct nfq_data *nfa,void *data)
{
	int id;				/* netfilter packet id		*/
	int len;			/* tptal datagram length	*/
	int ilen;			/* ip header length		*/
	int dlen;			/* tcp/udp total length		*/
	int doff;			/* tcp/udp payload offset	*/
	unsigned short mem;		/* authentication value memory	*/
	struct nfqnl_msg_packet_hdr *ph;/* netfilter packet header	*/
	union
	{
		struct iphdr *iph;	/* ip header			*/
		char *pkt;		/* ip header as char		*/
		unsigned char *bfr;	/* ip header as unsigned char	*/
	}u;
	union
	{
		unsigned char *bfr;	/* tcp/udp header as uns. char	*/
		struct udphdr *udp;	/* udp header			*/
		struct tcphdr *tcp;	/* tcp header			*/
	}v;
	struct
	{
		unsigned int timestamp;	/* replay protect timestamp or 0*/
		unsigned short id;	/* IP ID of datagram		*/
		unsigned short len;	/* datagram total length	*/
	}w;
	union
	{
		MD5(auth);		/* hmac output as bytes		*/
		unsigned short data[8];	/* hmac output as 16 bit values	*/
	}x;
	MD5DATA ctx;			/* hmac working context		*/


	/* try to get the netfilter packet id or use 0 if not possible */

	if((ph=nfq_get_msg_packet_hdr(nfa)))id=ntohl(ph->packet_id);
	else id=0;

	/* get packet, ignore packet if it is too short */

	if((len=nfq_get_payload(nfa,&u.pkt))<sizeof(struct iphdr)||
		len<(ilen=(u.iph->ihl<<2)))goto fail;

	/* ignore the packet if specified datagram length is larger than
	   the received packet length */

	if(len<ntohs(u.iph->tot_len))goto fail;

	/* ignore all fragmented packets */

	if(ntohs(u.iph->frag_off)&(IP_MF|IP_OFFMASK))goto fail;

	/* get pointer to protocol specific datagram part, get length
	   of protocol specific part */

	v.bfr=u.bfr+ilen;
	dlen=len-ilen;

	/* act according to protocol */

	switch(u.iph->protocol)
	{
	/* tcp */

	case IPPROTO_TCP:

		/* skip packet if protocol header is too short,
		   calculate payload offset from tcp header start */

		if(dlen<sizeof(struct tcphdr))goto fail;
		doff=v.tcp->doff<<2;
		if(dlen<doff)goto fail;

		/* if authentication is enabled */

		if(hmacmode)
		{
			/* memorize and reset tcp checksum */

			mem=v.tcp->check;
			v.tcp->check=0;

			/* start hmac over tcp datagram part */

			md5hminit(&ctx,&hmac2);
			md5hmnext(v.bfr,dlen,&ctx);
		}
		break;

	/* udp */

	case IPPROTO_UDP:

		/* skip packet if protocol header is too short */

		if(dlen<sizeof(struct udphdr)||dlen<ntohs(v.udp->len))goto fail;

		/* set up udp payload offset from udp header start */

		doff=sizeof(struct udphdr);

		/* if authentication is enabled */

		if(hmacmode)
		{
			/* memorize and reset udp checksum */

			mem=v.udp->check;
			v.udp->check=0;

			/* start hmac over udp datagram part */

			md5hminit(&ctx,&hmac2);
			md5hmnext(v.bfr,dlen,&ctx);
		}
		break;

	/* skip all other protocols */

	default:goto fail;
	}

	/* use zero timestamp if replay detection is disabled */

	if(!usetime)w.timestamp=0;

	/* if replay detection is enabled */

	else
	{
		/* setup local timestamp */

		w.timestamp=((unsigned int)time(NULL))>>8;

		/* get remote timestamp info */

		w.id=u.iph->tos>>6;

		/* adjust local timestamp according to remote info */

		if(w.id==(w.timestamp&3))goto timeok;
		else if((w.timestamp&3)==((w.id+1)&3))
			w.timestamp=(w.timestamp-1)&0xffffff;
		else if((w.timestamp&3)==((w.id-1)&3))
			w.timestamp=(w.timestamp+1)&0xffffff;
	}

	/* make timestamp endian independent */

timeok:	w.timestamp=htonl(w.timestamp);

	/* copy IP ID and IP datagram length */

	w.id=u.iph->id;
	w.len=u.iph->tot_len;

	/* decrypt payload */

	decrypt(&w,sizeof(w),v.bfr+doff,dlen-doff);

	/* act according to protocol */

	switch(u.iph->protocol)
	{
	/* tcp */

	case IPPROTO_TCP:

		/* if authentication is enabled */

		if(hmacmode)
		{
			/* add IV source data to hmac and complete hmac */

			md5hmnext((unsigned char *)&w,sizeof(w),&ctx);
			md5hmend(x.auth,&ctx,&hmac2);

			/* discard packet, if received checksum and first two
			   hmac bytes do not match */

			if(x.data[0]!=mem)goto discard;
		}

		/* done if no authentication or checksumming disabled */

		else if(!rxtcp)break;

		/* update tcp checksum */

		v.tcp->check=0;
		v.tcp->check=csum(u.iph,v.bfr,dlen);
		break;

	/* udp */

	case IPPROTO_UDP:

		/* if authentication is enabled */

		if(hmacmode)
		{
			/* add IV source data to hmac and complete hmac */

			md5hmnext((unsigned char *)&w,sizeof(w),&ctx);
			md5hmend(x.auth,&ctx,&hmac2);

			/* discard packet, if received checksum and first two
			   hmac bytes do not match */

			if(x.data[0]!=mem)goto discard;

			/* if udp checksum update is disabled */

			if(!rxudp)
			{
				/* mark udo checksum as not present */

				v.udp->check=0;

				/* done */

				break;
			}
		}

		/* done if no authentication, no checksum or checksumming
		   disabled */

		else if(!v.udp->check||!rxudp)break;

		/* update udp checksum */

		v.udp->check=0;
		v.udp->check=csum(u.iph,v.bfr,dlen);
		if(!v.udp->check)v.udp->check=0xffff;
		break;
	}

	/* in case of a set DF bit */

	if(ntohs(u.iph->frag_off)&IP_DF)
	{
		/* zero the IP ID field */

		u.iph->id=0;

		/* continue according to the replay protection mode */

		if(usetime)goto timecsum;
		goto ipcsum;
	}

	/* if replay protection is enabled */

	else if(usetime)
	{
		/* clear the two IP TOS MSB */

timecsum:	u.iph->tos&=0x3f;

		/* generate new ip checksum */

ipcsum:		u.iph->check=0;
		u.iph->check=ip_csum(u.bfr,ilen);
	}

	/* return modified packet to kernel */

	return nfq_set_verdict(qh,id,NF_ACCEPT,len,u.bfr);

	/* tell kernel to pass the unaltered packet */

fail:	return nfq_set_verdict(qh,id,NF_ACCEPT,0,NULL);

	/* tell kernel to discard the packet */

discard:return nfq_set_verdict(qh,id,NF_DROP,0,NULL);
}

/*
 * usage
 *
 * description:	This routine prints a usage message.
 */

static void usage(void)
{
fprintf(stderr,"\nUsage: ipcrypt <options>\n\n"
"-k <keyfile>    key file (default: " DFTKEY ")\n"
"-p <pidfile>    pid file (default: " DFTPID ")\n"
"-S <scriptfile> nfqueue setup script (default: " DFTSCR ")\n"
"-s <sendqueue>  nfqueue queue number for encryption (default: " DFTTX ")\n"
"-r <recvqueue>  nfqueue queue number for decryption (default: " DFTRX ")\n"
"-o <username>   run as this user\n"
"-g <groupname>  run as this group\n"
"-f              stay in foreground (don't daemonize)\n"
"-n              don't use realtime process priority\n"
"-R              use minimalistic replay prevention\n"
"-H              use minimalistic message authentication\n"
"-u              don't recalculate udp checksum after encrypt\n"
"-U              don't recalculate udp checksum after decrypt\n"
"-t              don't recalculate tcp checksum after encrypt\n"
"-T              don't recalculate tcp checksum after decrypt\n"
"-K              create specified or default key file\n"
"-h              display this help\n\n"
"Warning: ipcrypt is a minimalistic special purpose encryption solution.\n"
"Use e.g. openssh, OpenVPN or IPSec if you need a complete encryption solution."
"\n\n");
}

/*
 * mkkey
 *
 * return: 0 in case of success, -1 in case of error
 *
 * description:	This routine generates a key file for ipcrypt usage.
 */

static int mkkey(void)
{
	int fd;		/* file handle			*/
	int l;		/* current read length		*/
	int len;	/* current total bytes read	*/


	/* prevent key file acces for group and others */

	umask(0077);

	/* open random device, abort on error */

	if((fd=open(RANDOMDEV,O_RDONLY))==-1)return -1;

	/* read key data from random device, abort on error */

	for(len=0;len<KEYLENGTH;len+=l)
		if((l=read(fd,bfr+len,KEYLENGTH-len))<0)return -1;

	/* close random device */

	close(fd);

	/* open key file, abort on error */

	if((fd=open(keyfile,O_WRONLY|O_CREAT|O_TRUNC,0600))==-1)return -1;

	/* write key data to key file, abort on error */

	if(write(fd,bfr,KEYLENGTH)!=KEYLENGTH)return -1;

	/* close key file */

	close(fd);

	/* clear memory */

	memset(bfr,0,KEYLENGTH);

	/* signal success */

	return 0;
}

/*
 * die
 *
 * input: dummy - unused
 *
 * description:	This routine cleans up and terminates the ipcrypt process.
 *		It never returns to the caller.
 */

static void die(int dummy)
{
	/* call config script as root (remember, we have CAP_SETUID) */

	if(ipt)switch(fork())
	{
	case 0:	setresuid(0,0,0);
		setresgid(0,0,0);
		execl(script,script,"down",txqnam,rxqnam,NULL);
		exit(0);
	case -1:break;
	default:wait(NULL);
	}

	/* close netfilter socket handle,mark handle as closed */

	if(h)nfq_close(h);
	h=NULL;

	/* remove pid file */

	unlink(pidfile);

	/* clean up memory */

	memset(sched,0,sizeof(sched));
	memset(&hmac,0,sizeof(hmac));
	memset(&hmac2,0,sizeof(hmac2));
	memset(white,0,sizeof(white));

	/* release memory lock */

	munlockall();

	/* done */

	exit(1);
}

/*
 * main
 *
 * input:  argc - the argument vector count
 *	   argv - pointer to the argument vectors
 *
 * return: 1 in case of error
 *
 * description:	This routine is the main routine of the ipcrypt process.
 */

int main(int argc,char *argv[])
{
	int fd;					/* file descriptor	*/
	int len;				/* general purpose	*/
	int txq;				/* tx nfqueue number	*/
	int rxq;				/* rx nfqueue number	*/
	int mkfile=0;				/* make key file flag	*/
	int foreground=0;			/* no daemon flag	*/
	int norealtime=0;			/* no realtime flag	*/
	struct passwd *pw=NULL;			/* user to run as	*/
	struct group *gr=NULL;			/* group to run as	*/
	FILE *fp;				/* for pid file		*/
	struct timeval tv;			/* IP ID setup		*/
	struct sched_param sch;			/* realtime setup	*/
	struct stat stb;			/* key file sanity test	*/
	struct __user_cap_header_struct ch;	/* capability setup	*/
	struct __user_cap_data_struct cd;	/* capability setup	*/


	/* die an the following set of signals */

	signal(SIGHUP,die);
	signal(SIGINT,die);
	signal(SIGTERM,die);
	signal(SIGQUIT,die);
	signal(SIGSTOP,die);

	/* preset netfilter tx and rx queue numbers */

	txq=atoi(txqnam);
	rxq=atoi(rxqnam);

	/* set up initial local IP ID value */

	gettimeofday(&tv,NULL);
	seq=tv.tv_usec^tv.tv_sec;

	/* process all command line arguments */

	while((len=getopt(argc,argv,"k:p:s:r:S:o:g:fRuUtTnHKh"))!=-1)switch(len)
	{
	/* key file */

	case 'k':
		keyfile=optarg;
		break;

	/* pid file */

	case 'p':
		pidfile=optarg;
		break;

	/* netfilter tx queue number */

	case 's':
		if((txq=atoi(optarg))<=0)
		{
			fprintf(stderr,"Illegal tx queue value, aborting\n");
			die(0);
		}
		txqnam=optarg;
		break;

	/* net filter rx queue number */

	case 'r':
		if((rxq=atoi(optarg))<=0)
		{
			fprintf(stderr,"Illegal rx queue value, aborting\n");
			die(0);
		}
		rxqnam=optarg;
		break;

	/* netfilter config script */

	case 'S':
		script=optarg;
		break;

	/* foreground flag */

	case 'f':
		foreground=1;
		break;

	/* replay prevention flag */

	case 'R':
		usetime=1;
		break;

	/* no udp rx checksum update flag */

	case 'u':
		rxudp=0;
		break;

	/* no udp tx checksum uodate flag */

	case 'U':
		txudp=0;
		break;

	/* no tcp rx checksum update flag */

	case 't':
		rxtcp=0;
		break;

	/* no tcp tx checksum update flag */

	case 'T':
		txtcp=0;
		break;

	/* no realtime priority flag */

	case 'n':
		norealtime=1;
		break;

	/* authentication mode flag */

	case 'H':
		hmacmode=1;
		break;

	/* run as the specified user */

	case 'o':
		if(!(pw=getpwnam(optarg)))
		{
			fprintf(stderr,"No such user, aborting\n");
			die(0);
		}
		break;

	/* run as the specified group */

	case 'g':
		if(!(gr=getgrnam(optarg)))
		{
			fprintf(stderr,"No such group, aborting\n");
			die(0);
		}
		break;

	/* generate key file flag */

	case 'K':
		mkfile=1;
		break;

	/* print usage message */

	case 'h':
		usage();
		die(0);

	/* unknown option */

	default:fprintf(stderr,"Illegal option, aborting\n");
		die(0);
	}

	/* assert that we are root */

	if(geteuid())
	{
		fprintf(stderr,"Not root, aborting.\n");
		die(0);
	}

	/* if a key file has to be generated */

	if(mkfile)
	{
		/* try to create key file, handle errors */

		if(mkkey())
		{
			fprintf(stderr,"Can't create key file, aborting.\n");
			die(0);
		}

		/* notify user */

		fprintf(stderr,"Key file created, done.\n");

		/* done */

		return 0;
	}

	/* open key file abort on error */

	if((fd=open(keyfile,O_RDONLY))==-1)
	{
		fprintf(stderr,"No key file, aborting.\n");
		die(0);
	}

	/* get key file info, abort on error */

	if(fstat(fd,&stb))
	{
		close(fd);
		fprintf(stderr,"Key file metadata not accessible, aborting.\n");
		die(0);
	}

	/* abort on key file security error */

	if((stb.st_mode&0077)||stb.st_uid||stb.st_gid)
	{
		close(fd);
		fprintf(stderr,"Insecure key file, aborting.\n");
		die(0);
	}

	/* read key file */

	if(read(fd,bfr,KEYLENGTH)!=KEYLENGTH)
	{
		close(fd);
		memset(bfr,0,KEYLENGTH);
		fprintf(stderr,"Key file too short, aborting.\n");
		die(0);
	}

	/* close key file */

	close(fd);

	/* set up hmac key for IV generation */

	md5hmkey((unsigned char *)bfr,64,&hmac);

	/* set up hmac key for authentication */

	md5hmkey((unsigned char *)bfr+64,64,&hmac2);

	/* set up whitening key schedule */

	blowfish_key_schedule((unsigned char *)bfr+128,64,sched);

	/* preset whitening buffer with well known values */


	for(len=0;len<256;len++)white[len]=(WORD08)(len);

	/* encrypt whitening buffer */

	for(len=0;len<256;len+=BLOWFISH_BLOCKSIZE)
	{
		blowfish_encrypt(white+len,sched);
		blowfish_encrypt(white+len,sched);
		blowfish_encrypt(white+len,sched);
		blowfish_encrypt(white+len,sched);
	}

	/* set up encryption/decryption key schedule */

	blowfish_key_schedule((unsigned char *)bfr+192,64,sched);

	/* remove keys from memory */

	memset(bfr,0,KEYLENGTH);

	/* nice ourselves, abort on error */

	errno=0;
	nice(-20);
	if(errno)
	{
		fprintf(stderr,"Priority setup failed, aborting.\n");
		die(0);
	}

	/* lock us into memory,m abort on error */

	if(mlockall(MCL_CURRENT|MCL_FUTURE))
	{
		fprintf(stderr,"Memory locking failed, aborting.\n");
		die(0);
	}

	/* if realtime processing is enabled */

	if(!norealtime)
	{
		/* set up realtime priority 99, abort on error */

		memset(&sch,0,sizeof(sch));
		sch.sched_priority=99;
		if(sched_setscheduler(0,SCHED_RR,&sch))
		{
			fprintf(stderr,"Scheduler setup failed, aborting.\n");
			die(0);
		}
	}

	/* open netfilter handle, abort on error */

	if((h=nfq_open())==NULL)
	{
		fprintf(stderr,"No nfqueue handle, aborting.\n");
		die(0);
	}

	/* ask the netfilter folks why one has to call unbind... */

	if(nfq_unbind_pf(h,AF_INET)<0)
	{
		fprintf(stderr,"No nfqueue unbind, aborting.\n");
		die(0);
	}

	/* bind ourselves to IPv4 */

	if(nfq_bind_pf(h,AF_INET)<0)
	{
		fprintf(stderr,"No nfqueue bind, aborting.\n");
		die(0);
	}

	/* create tx netfilter queue, abort on error */

	if(!(qh1=nfq_create_queue(h,txq,&tx,NULL)))
	{
		fprintf(stderr,"No nfqueue tx queue, aborting.\n");
		die(0);
	}

	/* set queue to copy the entire packet */

	if(nfq_set_mode(qh1,NFQNL_COPY_PACKET,0xffff)<0)
	{
		fprintf(stderr,"nfqueue tx mode setup failed, aborting.\n");
		die(0);
	}

	/* create rx netfilter queue, abort on error */

	if(!(qh2=nfq_create_queue(h,rxq,&rx,NULL)))
	{
		fprintf(stderr,"No nfqueue rx queue, aborting.\n");
		die(0);
	}

	/* set queue to copy the entire packet */

	if(nfq_set_mode(qh2,NFQNL_COPY_PACKET,0xffff)<0)
	{
		fprintf(stderr,"nfqueue rx mode setup failed, aborting.\n");
		die(0);
	}

	/* daemonize if not foreground processing enabled */

	if(!foreground)if(daemon(0,0))
	{
		fprintf(stderr,"Cannot daemonize, aborting.\n");
		die(0);
	}

	/* set up our group id, if requested */

	if(gr)setresgid(gr->gr_gid,gr->gr_gid,gr->gr_gid);

	/* if a user id was requested */

	if(pw)
	{
		/* set up required capabilities */

		memset(&ch,0,sizeof(ch));
		ch.version=_LINUX_CAPABILITY_VERSION;
		ch.pid=0;
		memset(&cd,0,sizeof(cd));
		cd.effective=1<<CAP_NET_RAW|1<<CAP_NET_ADMIN|1<<CAP_SETUID;
		cd.permitted=1<<CAP_NET_RAW|1<<CAP_NET_ADMIN|1<<CAP_SETUID;
		cd.inheritable=1;
		capset(&ch,&cd);

		/* keep capabilities over next call */

		prctl(PR_SET_KEEPCAPS,1,0,0,0);

		/* change user id */

		setresuid(pw->pw_uid,pw->pw_uid,pw->pw_uid);

		/* set up again required capabilities */

		memset(&ch,0,sizeof(ch));
		ch.version=_LINUX_CAPABILITY_VERSION;
		ch.pid=0;
		memset(&cd,0,sizeof(cd));
		cd.effective=1<<CAP_NET_RAW|1<<CAP_NET_ADMIN|1<<CAP_SETUID;
		cd.permitted=1<<CAP_NET_RAW|1<<CAP_NET_ADMIN|1<<CAP_SETUID;
		cd.inheritable=1;
		capset(&ch,&cd);

		/* don't keep capabilities when changing user id */

		prctl(PR_SET_KEEPCAPS,0,0,0,0);
	}

	/* try to write pid to file, ignore errors */

	if((fp=fopen(pidfile,"w")))
	{
		fprintf(fp,"%d\n",getpid());
		fclose(fp);
	}

	/* call config script as root (remember, we have CAP_SETUID) */

	switch(fork())
	{
	case 0:	setresuid(0,0,0);
		setresgid(0,0,0);
		execl(script,script,"up",txqnam,rxqnam,NULL);
		exit(0);
	case -1:break;
	default:wait(NULL);
		ipt=1;
	}

	/* get socket handle from netfilter descriptor */

	fd=nfq_fd(h);

	/* do, until signalled */

	while(1)
	{
		/* receive next netfilter packet */

		if((len=recv(fd,bfr,sizeof(bfr),0))<0)continue;

		/* process netfilter packet */

		nfq_handle_packet(h,bfr,len);
	}
}
